/*
* This file is part of HOBETA2TRD.
*
* Copyright (C) 2025 zxcygnus <cygnus1 at post.cz>
* Modifications (C) 2015 ub880d <ub880d@users.sf.net> (mainly for pedantic compile)
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or (at
* your option) any later version.
*
* This program is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
* General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
* USA
*/

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>		// strcpy
#include <stdint.h>		// integer types
#include "trdos_structure.h"

#define VERSION		"0.3"
#define	DEBUG		0

/* Hobeta head offsets - nelze jinak, protože kompiler GCC zarovnává na 2 byty a to se kvůli 9 bytovému názvu silně nehodí */
#define HH_FILENAME	0
#define HH_EXTENSION	8
#define HH_ADDRESS	9
#define HH_LENGTH	11
#define HH_TR		13
#define HH_SECTORS	14
#define HH_CRC		15

/* TRDOS head offsets in directory of disk image */
#define TR_FILENAME	0
#define TR_EXTENSION	8
#define TR_ADDRESS	9
#define TR_LENGTH	11
#define TR_SECTORS	13
#define TR_SECTOR	14
#define TR_TRACK	15

/* structure needed for reading whole modified TRDOS header from hobeta file */
typedef struct {
	unsigned char byte[17];
} hobetahead;

/* structure needed for processing start line of BASIC program */
typedef struct {
	unsigned char byte[4];
} fourbytes;


/* global options */
int	verbose = 0;
int	force = 0;
int	colors = 0;
int	dryrun = 0;
int	errno;

/* selection of ANSI colors */
#define CRED     "\x1b[31m"
#define CGREEN   "\x1b[32m"
#define CYELLOW  "\x1b[33m"
#define CBLUE    "\x1b[34m"
#define CMAGEN   "\x1b[35m"
#define CCYAN    "\x1b[36m"
#define CBRED    "\x1b[91m"
#define CBGREEN  "\x1b[92m"
#define CBYELLOW "\x1b[93m"
#define CBWHITE  "\x1b[97m"
#define CRESET   "\x1b[0m"

/* CRC-32 (Ethernet, ZIP, etc.) polynomial in reversed bit order. */
#define	POLY	0xedb88320

typedef enum {none, black, red, green, yellow, white, brightred, brightgreen, brightyellow, brightwhite} listofcolors_t;


uint32_t crc32Lookup[256] = {
	0x00000000,0x77073096,0xEE0E612C,0x990951BA,0x076DC419,0x706AF48F,0xE963A535,0x9E6495A3,
	0x0EDB8832,0x79DCB8A4,0xE0D5E91E,0x97D2D988,0x09B64C2B,0x7EB17CBD,0xE7B82D07,0x90BF1D91,
	0x1DB71064,0x6AB020F2,0xF3B97148,0x84BE41DE,0x1ADAD47D,0x6DDDE4EB,0xF4D4B551,0x83D385C7,
	0x136C9856,0x646BA8C0,0xFD62F97A,0x8A65C9EC,0x14015C4F,0x63066CD9,0xFA0F3D63,0x8D080DF5,
	0x3B6E20C8,0x4C69105E,0xD56041E4,0xA2677172,0x3C03E4D1,0x4B04D447,0xD20D85FD,0xA50AB56B,
	0x35B5A8FA,0x42B2986C,0xDBBBC9D6,0xACBCF940,0x32D86CE3,0x45DF5C75,0xDCD60DCF,0xABD13D59,
	0x26D930AC,0x51DE003A,0xC8D75180,0xBFD06116,0x21B4F4B5,0x56B3C423,0xCFBA9599,0xB8BDA50F,
	0x2802B89E,0x5F058808,0xC60CD9B2,0xB10BE924,0x2F6F7C87,0x58684C11,0xC1611DAB,0xB6662D3D,
	0x76DC4190,0x01DB7106,0x98D220BC,0xEFD5102A,0x71B18589,0x06B6B51F,0x9FBFE4A5,0xE8B8D433,
	0x7807C9A2,0x0F00F934,0x9609A88E,0xE10E9818,0x7F6A0DBB,0x086D3D2D,0x91646C97,0xE6635C01,
	0x6B6B51F4,0x1C6C6162,0x856530D8,0xF262004E,0x6C0695ED,0x1B01A57B,0x8208F4C1,0xF50FC457,
	0x65B0D9C6,0x12B7E950,0x8BBEB8EA,0xFCB9887C,0x62DD1DDF,0x15DA2D49,0x8CD37CF3,0xFBD44C65,
	0x4DB26158,0x3AB551CE,0xA3BC0074,0xD4BB30E2,0x4ADFA541,0x3DD895D7,0xA4D1C46D,0xD3D6F4FB,
	0x4369E96A,0x346ED9FC,0xAD678846,0xDA60B8D0,0x44042D73,0x33031DE5,0xAA0A4C5F,0xDD0D7CC9,
	0x5005713C,0x270241AA,0xBE0B1010,0xC90C2086,0x5768B525,0x206F85B3,0xB966D409,0xCE61E49F,
	0x5EDEF90E,0x29D9C998,0xB0D09822,0xC7D7A8B4,0x59B33D17,0x2EB40D81,0xB7BD5C3B,0xC0BA6CAD,
	0xEDB88320,0x9ABFB3B6,0x03B6E20C,0x74B1D29A,0xEAD54739,0x9DD277AF,0x04DB2615,0x73DC1683,
	0xE3630B12,0x94643B84,0x0D6D6A3E,0x7A6A5AA8,0xE40ECF0B,0x9309FF9D,0x0A00AE27,0x7D079EB1,
	0xF00F9344,0x8708A3D2,0x1E01F268,0x6906C2FE,0xF762575D,0x806567CB,0x196C3671,0x6E6B06E7,
	0xFED41B76,0x89D32BE0,0x10DA7A5A,0x67DD4ACC,0xF9B9DF6F,0x8EBEEFF9,0x17B7BE43,0x60B08ED5,
	0xD6D6A3E8,0xA1D1937E,0x38D8C2C4,0x4FDFF252,0xD1BB67F1,0xA6BC5767,0x3FB506DD,0x48B2364B,
	0xD80D2BDA,0xAF0A1B4C,0x36034AF6,0x41047A60,0xDF60EFC3,0xA867DF55,0x316E8EEF,0x4669BE79,
	0xCB61B38C,0xBC66831A,0x256FD2A0,0x5268E236,0xCC0C7795,0xBB0B4703,0x220216B9,0x5505262F,
	0xC5BA3BBE,0xB2BD0B28,0x2BB45A92,0x5CB36A04,0xC2D7FFA7,0xB5D0CF31,0x2CD99E8B,0x5BDEAE1D,
	0x9B64C2B0,0xEC63F226,0x756AA39C,0x026D930A,0x9C0906A9,0xEB0E363F,0x72076785,0x05005713,
	0x95BF4A82,0xE2B87A14,0x7BB12BAE,0x0CB61B38,0x92D28E9B,0xE5D5BE0D,0x7CDCEFB7,0x0BDBDF21,
	0x86D3D2D4,0xF1D4E242,0x68DDB3F8,0x1FDA836E,0x81BE16CD,0xF6B9265B,0x6FB077E1,0x18B74777,
	0x88085AE6,0xFF0F6A70,0x66063BCA,0x11010B5C,0x8F659EFF,0xF862AE69,0x616BFFD3,0x166CCF45,
	0xA00AE278,0xD70DD2EE,0x4E048354,0x3903B3C2,0xA7672661,0xD06016F7,0x4969474D,0x3E6E77DB,
	0xAED16A4A,0xD9D65ADC,0x40DF0B66,0x37D83BF0,0xA9BCAE53,0xDEBB9EC5,0x47B2CF7F,0x30B5FFE9,
	0xBDBDF21C,0xCABAC28A,0x53B39330,0x24B4A3A6,0xBAD03605,0xCDD70693,0x54DE5729,0x23D967BF,
	0xB3667A2E,0xC4614AB8,0x5D681B02,0x2A6F2B94,0xB40BBE37,0xC30C8EA1,0x5A05DF1B,0x2D02EF8D
	};


void printzxchars (unsigned char c)
{
	/* This function converts ZX Spectrum characters for Linux console as UTF-8 */
	switch (c)
	{
		/* symbol for unknown UTF-8 character should be */
		/* U+FFFD question mark in black square		*/
		/* U+25A1 white square 				*/
		case 0 ... 31:
			printf ("□");		/* control codes will be printed as UTF-8 unknown character white square */
			break;
		case 32 ... 95:
			printf ("%c", c);	/* print directly as is */
			break;
		case 96:
			printf ("£");		/* british pound */
			break;
		case 97 ... 126:
			printf ("%c", c);	/* print directly as is */
			break;
		case 127:
			printf ("©");		/* copyright */
			break;
		case 128:
			printf (" ");		/* GRAPH - empty box*/
			break;
		case 129:
			printf ("▝");		/* GRAPH - quadrant upper right */
			break;
		case 130:
			printf ("▘");		/* GRAPH - quadrant upper left */
			break;
		case 131:
			printf ("▀");		/* GRAPH - upper half block */
			break;
		case 132:
			printf ("▗");		/* GRAPH - quadrant lower right */
			break;
		case 133:
			printf ("▐");		/* GRAPH - light half block */
			break;
		case 134:
			printf ("▚");		/* GRAPH - quadrant upper left and lower righr */
			break;
		case 135:
			printf ("▜");		/* GRAPH - quadrant upper left and upper right and lower right */
			break;
		case 136:
			printf ("▖");		/* GRAPH - quadrant lower left */
			break;
		case 137:
			printf ("▞");		/* GRAPH - quadrant upper right and lower left */
			break;
		case 138:
			printf ("▌");		/* GRAPH - left half block */
			break;
		case 139:
			printf ("▛");		/* GRAPH - quadrant upper left and upper right and lower left */
			break;
		case 140:
			printf ("▄");		/* GRAPH - lower half block */
			break;
		case 141:
			printf ("▟");		/* GRAPH - quadrant upper right and lower left and lower right */
			break;
		case 142:
			printf ("▙");		/* GRAPH - quadrant upper left and lower left and lower right */
			break;
		case 143:
			printf ("█");		/* GRAPH - full block */
			break;
		case 144 ... 164:
			printf ("%c", c-79);	/* GRAPH - same as ASCII characters 'A'...'U' */
			break;
		case 165:
			printf ("RND");		/* TOKEN - BASIC command */
			break;
		case 166:
			printf ("INKEY$");	/* TOKEN - BASIC command */
			break;
		case 167:
			printf ("PI");		/* TOKEN - BASIC command */
			break;
		case 168:
			printf ("FN ");		/* TOKEN - BASIC command */
			break;
		case 169:
			printf ("POINT ");	/* TOKEN - BASIC command */
			break;
		case 170:
			printf ("SCREEN$ ");	/* TOKEN - BASIC command */
			break;
		case 171:
			printf ("ATTR ");	/* TOKEN - BASIC command */
			break;
		case 172:
			printf ("AT ");		/* TOKEN - BASIC command */
			break;
		case 173:
			printf ("TAB ");	/* TOKEN - BASIC command */
			break;
		case 174:
			printf ("VAL$ ");	/* TOKEN - BASIC command */
			break;
		case 175:
			printf ("CODE ");	/* TOKEN - BASIC command */
			break;
		case 176:
			printf ("VAL ");	/* TOKEN - BASIC command */
			break;
		case 177:
			printf ("LEN ");	/* TOKEN - BASIC command */
			break;
		case 178:
			printf ("SIN ");	/* TOKEN - BASIC command */
			break;
		case 179:
			printf ("COS ");	/* TOKEN - BASIC command */
			break;
		case 180:
			printf ("TAN ");	/* TOKEN - BASIC command */
			break;
		case 181:
			printf ("ASN ");	/* TOKEN - BASIC command */
			break;
		case 182:
			printf ("ACS ");	/* TOKEN - BASIC command */
			break;
		case 183:
			printf ("ATN ");	/* TOKEN - BASIC command */
			break;
		case 184:
			printf ("LN ");		/* TOKEN - BASIC command */
			break;
		case 185:
			printf ("EXP ");	/* TOKEN - BASIC command */
			break;
		case 186:
			printf ("INT ");	/* TOKEN - BASIC command */
			break;
		case 187:
			printf ("SQR ");	/* TOKEN - BASIC command */
			break;
		case 188:
			printf ("SGN ");	/* TOKEN - BASIC command */
			break;
		case 189:
			printf ("ABS ");	/* TOKEN - BASIC command */
			break;
		case 190:
			printf ("PEEK ");	/* TOKEN - BASIC command */
			break;
		case 191:
			printf ("IN ");		/* TOKEN - BASIC command */
			break;
		case 192:
			printf ("USR ");	/* TOKEN - BASIC command */
			break;
		case 193:
			printf ("STR$ ");	/* TOKEN - BASIC command */
			break;
		case 194:
			printf ("CHR$ ");	/* TOKEN - BASIC command */
			break;
		case 195:
			printf ("NOT ");	/* TOKEN - BASIC command */
			break;
		case 196:
			printf ("BIN ");	/* TOKEN - BASIC command */
			break;
		case 197:
			printf ("OR ");		/* TOKEN - BASIC command */
			break;
		case 198:
			printf ("AND ");	/* TOKEN - BASIC command */
			break;
		case 199:
			printf ("<=");		/* TOKEN - BASIC command */
			break;
		case 200:
			printf (">=");		/* TOKEN - BASIC command */
			break;
		case 201:
			printf ("<>");		/* TOKEN - BASIC command */
			break;
		case 202:
			printf ("LINE ");	/* TOKEN - BASIC command */
			break;
		case 203:
			printf ("THEN ");	/* TOKEN - BASIC command */
			break;
		case 204:
			printf ("TO ");		/* TOKEN - BASIC command */
			break;
		case 205:
			printf ("STEP ");	/* TOKEN - BASIC command */
			break;
		case 206:
			printf ("DEF FN ");	/* TOKEN - BASIC command */
			break;
		case 207:
			printf ("CAT ");	/* TOKEN - BASIC command */
			break;
		case 208:
			printf ("FORMAT ");	/* TOKEN - BASIC command */
			break;
		case 209:
			printf ("MOVE ");	/* TOKEN - BASIC command */
			break;
		case 210:
			printf ("ERASE ");	/* TOKEN - BASIC command */
			break;
		case 211:
			printf ("OPEN #");	/* TOKEN - BASIC command */
			break;
		case 212:
			printf ("CLOSE #");	/* TOKEN - BASIC command */
			break;
		case 213:
			printf ("MERGE ");	/* TOKEN - BASIC command */
			break;
		case 214:
			printf ("VERIFY ");	/* TOKEN - BASIC command */
			break;
		case 215:
			printf ("BEEP ");	/* TOKEN - BASIC command */
			break;
		case 216:
			printf ("CIRCLE ");	/* TOKEN - BASIC command */
			break;
		case 217:
			printf ("INK ");	/* TOKEN - BASIC command */
			break;
		case 218:
			printf ("PAPER ");	/* TOKEN - BASIC command */
			break;
		case 219:
			printf ("FLASH ");	/* TOKEN - BASIC command */
			break;
		case 220:
			printf ("BRIGHT ");	/* TOKEN - BASIC command */
			break;
		case 221:
			printf ("INVERSE ");	/* TOKEN - BASIC command */
			break;
		case 222:
			printf ("OVER ");	/* TOKEN - BASIC command */
			break;
		case 223:
			printf ("OUT ");	/* TOKEN - BASIC command */
			break;
		case 224:
			printf ("LPRINT ");	/* TOKEN - BASIC command */
			break;
		case 225:
			printf ("LLIST ");	/* TOKEN - BASIC command */
			break;
		case 226:
			printf ("STOP ");	/* TOKEN - BASIC command */
			break;
		case 227:
			printf ("READ ");	/* TOKEN - BASIC command */
			break;
		case 228:
			printf ("DATA ");	/* TOKEN - BASIC command */
			break;
		case 229:
			printf ("RESTORE ");	/* TOKEN - BASIC command */
			break;
		case 230:
			printf ("NEW ");	/* TOKEN - BASIC command */
			break;
		case 231:
			printf ("BORDER ");	/* TOKEN - BASIC command */
			break;
		case 232:
			printf ("CONTINUE ");	/* TOKEN - BASIC command */
			break;
		case 233:
			printf ("DIM ");	/* TOKEN - BASIC command */
			break;
		case 234:
			printf ("REM ");	/* TOKEN - BASIC command */
			break;
		case 235:
			printf ("FOR ");	/* TOKEN - BASIC command */
			break;
		case 236:
			printf ("GO TO ");	/* TOKEN - BASIC command */
			break;
		case 237:
			printf ("GO SUB ");	/* TOKEN - BASIC command */
			break;
		case 238:
			printf ("INPUT ");	/* TOKEN - BASIC command */
			break;
		case 239:
			printf ("LOAD ");	/* TOKEN - BASIC command */
			break;
		case 240:
			printf ("LIST ");	/* TOKEN - BASIC command */
			break;
		case 241:
			printf ("LET ");	/* TOKEN - BASIC command */
			break;
		case 242:
			printf ("PAUSE ");	/* TOKEN - BASIC command */
			break;
		case 243:
			printf ("NEXT ");	/* TOKEN - BASIC command */
			break;
		case 244:
			printf ("POKE ");	/* TOKEN - BASIC command */
			break;
		case 245:
			printf ("PRINT ");	/* TOKEN - BASIC command */
			break;
		case 246:
			printf ("PLOT ");	/* TOKEN - BASIC command */
			break;
		case 247:
			printf ("RUN ");	/* TOKEN - BASIC command */
			break;
		case 248:
			printf ("SAVE ");	/* TOKEN - BASIC command */
			break;
		case 249:
			printf ("RANDOMIZE ");	/* TOKEN - BASIC command */
			break;
		case 250:
			printf ("IF ");		/* TOKEN - BASIC command */
			break;
		case 251:
			printf ("CLS ");	/* TOKEN - BASIC command */
			break;
		case 252:
			printf ("DRAW ");	/* TOKEN - BASIC command */
			break;
		case 253:
			printf ("CLEAR ");	/* TOKEN - BASIC command */
			break;
		case 254:
			printf ("RETURN ");	/* TOKEN - BASIC command */
			break;
		case 255:
			printf ("COPY ");	/* TOKEN - BASIC command */
			break;
	}
}

/* separated in function, because this will be different for different operating systems */
void colorize (listofcolors_t c)
{
	/* only if colors are globally enabled */
	if (colors)
	{
		/* TODO: different colors for terminals with white background */
		switch (c)
		{
			case red:
				printf ("%s", CRED);
				break;
			case green:
				printf ("%s", CGREEN);
				break;
			case brightgreen:
				printf ("%s", CBGREEN);
				break;
			case brightyellow:
				printf ("%s", CBYELLOW);
				break;
			case brightwhite:
				printf ("%s", CBWHITE);
				break;
			default:
				printf ("%s", CRESET);
				break;
		}
	}
}

void colored_warning (void)
{
	colorize (brightyellow);
	printf ("Warning: ");
	colorize (none);
}

void colored_error (void)
{
	colorize (red);
	printf ("Error! ");
	colorize (none);
}

/* ... from readme included with program Hobeta

unsigned int CheckSum;
int i;

CheckSum=0;
for (i=0; i<=14; CheckSum = CheckSum + (header.head[i] * 257) + i, i++);
header.tr_head.tr_crc = CheckSum; */

int hobetainfo (char* source_filename)
{
	unsigned int i;
	unsigned char c;
	hobetahead header;
	fourbytes bytesbasicend;
	FILE *inputfile;
	const char *trdfiletype;
	unsigned int trdfilelengths;
	unsigned int trdfilelength;
	unsigned int trdfileaddress;
	int16_t hobetachecksum;
	int16_t basicstartline;

	inputfile = fopen (source_filename,"r");
	
	if (inputfile == NULL)
	{
		colored_error ();
		printf ("Can't open input file %s.\n", source_filename);
		exit (EXIT_FAILURE);
	}
	
	if (fread (&header, sizeof (hobetahead), 1, inputfile) == 1)
	{
		printf ("filename:\t");
		colorize (brightwhite);
		for (i = 0; i < 8; i++) printzxchars(header.byte[HH_FILENAME+i]);
		printf (".");
		printzxchars(header.byte[HH_EXTENSION]);
		colorize (none);
		printf ("\n");
	}
	
	/* description of filetype only in verbose mode, extension is mostly enough */
	if (verbose)
	{
		switch (header.byte[HH_EXTENSION])
		{
			case 'b':
			case 'B':
				trdfiletype = "BASIC PROGRAM";
				break;
			case 'c':
			case 'C':
				trdfiletype = "CODE (BYTES)";
				break;
			case 'd':
			case 'D':
				trdfiletype = "DATA";
				break;
			case 'M':
				trdfiletype = "SOUNDTRACKER";
				break;
			case '#':
				trdfiletype = "STREAM";
				break;
			default:
				trdfiletype = "UNKNOWN";
				break;
		}
		printf ("filetype:\t%s\n", trdfiletype);
	}
	
	/* length in sectors */
	c = header.byte[HH_SECTORS];
	printf ("sectors:\t");
	colorize (brightwhite);
	printf ("%hhu", c);
	colorize (none);
	if (verbose) printf ("\t0x%02x", c);
	printf ("\n");
	
	/* length in bytes, length without variables and start address of binary files */
	if (header.byte[HH_EXTENSION] == 'b' || header.byte[HH_EXTENSION] == 'B')
	{
		/* BASIC program have two lengths, but no start address in TRDOS header*/
		/* length without variables */
		i = header.byte[HH_LENGTH] + 256*header.byte[HH_LENGTH+1];
		printf ("length: \t");
		colorize (brightwhite);
		printf ("%hu", i);
		colorize (none);
		if (verbose) printf ("\t0x%04x", 0xffff & i);
		printf ("\n");
		
		/* length with variables (may be bigger) */
		i = header.byte[HH_ADDRESS] + 256*header.byte[HH_ADDRESS+1];
		printf ("length+var:\t");
		colorize (brightwhite);
		printf ("%hu", i);
		colorize (none);
		if (verbose) printf ("\t0x%04x", 0xffff & i);
		printf ("\n");
		
		/* read and decode BASIC start line */
		/* start line should be stored 2 bytes after length of BASIC program with variables */
		/* bytes between BASIC program and START LINE are 128, 170 (can there be different values?) */
		/* if START LINE is 0, that means program will not start automatically */
		
		fseek (inputfile, 17+i, SEEK_SET);
	
		if (fread (&bytesbasicend, sizeof (bytesbasicend), 1, inputfile) == 1)
		{
			if (DEBUG)
			{
				/* check two bytes before start line, should be  128, 170 */
				printf ("byte 128 = %hhu, ", bytesbasicend.byte[0]);
				printf ("byte 170 = %hhu, ", bytesbasicend.byte[1]);
			}
			
			basicstartline = bytesbasicend.byte[2] + 256*bytesbasicend.byte[3];
			
			if (basicstartline > 0)
			{
				printf ("startline:\t");
				colorize (brightwhite);
				printf ("%hu", basicstartline);
				colorize (none);
				if (verbose)
				{
					printf ("\t0x%04x", 0xffff & basicstartline);
					
					if (bytesbasicend.byte[0] == 128 && bytesbasicend.byte[1] == 170)
					{
						colorize (brightgreen);
						printf ("\tvalues 128,170 match");
						colorize (none);
					}
					else
					{
						colorize (red);
						printf ("\tvalues 128,170 are different!");
						colorize (none);
						printf (" Could be damaged content of Hobeta file?");
					}
				}
				printf ("\n");
			}
		}
		else
		{
			colored_error ();
			printf ("Reading from file %s failed.\n", source_filename);
			fclose (inputfile);
			exit (EXIT_FAILURE);
		}
	}
	else
	{
		/* all binary files have start address and length in bytes */
		i = header.byte[HH_ADDRESS] + 256*header.byte[HH_ADDRESS+1];
		printf ("start:  \t");
		colorize (brightwhite);
		printf ("%hu", i);
		colorize (none);
		if (verbose) printf ("\t0x%04x", 0xffff & i);
		printf ("\n");
		
		i = header.byte[HH_LENGTH] + 256*header.byte[HH_LENGTH+1];
		printf ("length: \t");
		colorize (brightwhite);
		printf ("%hu", i);
		colorize (none);
		if (verbose) printf ("\t0x%04x", 0xffff & i);
		printf ("\n");
	}
	
	if (verbose)
	{
		/* compute header checksum */
		hobetachecksum = 0;
		for (i = 0; i <= 14; hobetachecksum = hobetachecksum + (header.byte[i] * 257) + i, i++);
		
		printf ("checksum:\t%hu", hobetachecksum);
		printf ("\t0x%04x", 0xffff & hobetachecksum);
		
		i = header.byte[HH_CRC] + 256*header.byte[HH_CRC+1];
		if ((int16_t)i == (int16_t)hobetachecksum)
		{
			colorize (green);
			printf ("\tok");
			colorize (none);
		}
		else
		{
			colorize (red);
			printf ("\tdoesn't match!");
			colorize (none);
			printf ("\t%hu\t0x%04x", i, (0xffff & i));
		}
		printf ("\n");
	}
	
	return (EXIT_SUCCESS);
}


int convert (char* source_filename, char* target_filename)
{
	int i;
	int trdosfile;		/* counter for processed filename in TRDOS directory */
	uint8_t s;		/* variable for searching filenames - sector number */
	uint8_t o;		/* variable for searching filenames - offset in sector */
	int dupfound;		/* variable for searching filenames - will be 1 if duplicate file found in TRDOS image */
	int different;		/* variable for searching filenames - will be 1 if something in compared file differs */
	hobetahead header;
	FILE *inputfile;
	FILE *outputfile;
	trdos_sector sector;
	trdos_track track;
	unsigned int sector_number;
	unsigned int track_number;
	unsigned int freesectors;
	unsigned int directory_offset;
	unsigned int fatal_error_found;
	unsigned int dir_sector;
	int16_t hobetachecksum;

	inputfile = fopen (source_filename, "r");
	if (inputfile == NULL)
	{
		colored_error ();
		printf ("Can't open input file %s.\n", source_filename);
		exit (EXIT_FAILURE);
	}

	outputfile = fopen (target_filename, "ra+");
	if (outputfile == NULL)
	{
		colored_error ();
		printf ("Can't open output file %s.\n", target_filename);
		exit (EXIT_FAILURE);
	}

	/* read header from Hobeta file */
	if (fread (&header, sizeof (hobetahead), 1, inputfile) == 1)
	{
		/* compute header checksum */
		hobetachecksum = 0;
		for (i = 0; i <= 14; hobetachecksum = hobetachecksum + (header.byte[i] * 257) + i, i++);

		/* read whole track from TRDOS disk image - it contains whole directory and system sector... */
		if (fread (&track, sizeof (trdos_track), 1, outputfile) == 1)
		{	
			/* check if both files are correct as much as possible */
			fatal_error_found = 0;
			
			/* TRDOS image must contain TRDOS identification byte in system sector */
			if (track.sector[INFO_SEC].byte[OFFSET_TRDOSIDENT] != 16)
			{
				if (!force)
				{
					colored_warning ();
					printf ("Missing TRDOS identification in target image file! You can try -f (force) parameter, but i strongly recommend to backup target file and check again if it really is TRDOS image. \n");
					fatal_error_found = 1;
				}
				else
				{
					if (verbose) printf ("Missing TRDOS identification, but ignoring it, because option --force was used.\n");
				}
			}
			
			/* TRDOS image must have specified formated capacity - only 4 values are allowed */
			if (track.sector[INFO_SEC].byte[OFFSET_DISKFORMAT] != 22 && track.sector[INFO_SEC].byte[OFFSET_DISKFORMAT] != 23 && track.sector[INFO_SEC].byte[OFFSET_DISKFORMAT] != 24 && track.sector[INFO_SEC].byte[OFFSET_DISKFORMAT] != 25)
			{
				if (!force)
				{
					colored_warning ();
					printf ("Value of formated capacity is incorrect %hhu, should be 22, 23, 24, or 25.\n", track.sector[INFO_SEC].byte[OFFSET_DISKFORMAT]);
					fatal_error_found = 1;
				}
				else
				{
					if (verbose) printf ("Value of formated capacity is incorrect, but ignoring it, because option --force was used.\n");
				}
			}
			
			/* check number of files on TRDOS image, must be in range 0..128 because 16*128=2048 bytes */
			if (track.sector[INFO_SEC].byte[OFFSET_FILES] > 128)
			{
				if (!force)
				{
					colored_warning ();
					printf ("Number of files must be in range from 0 to 128, but is %hhu.\n", track.sector[INFO_SEC].byte[OFFSET_FILES]);
					fatal_error_found = 1;
				}
				else
				{
					if (verbose) printf ("Number of files isn't in range from 0 to 128, but ignoring it, because option --force was used.\n");
				}
			}
			
			/* check first free sector, must be in range 0..15 */
			if (track.sector[INFO_SEC].byte[OFFSET_FIRSTSECTOR] > 15 || track.sector[INFO_SEC].byte[OFFSET_FIRSTSECTOR] < 0)
			{
				if (!force)
				{
					colored_warning ();
					printf ("First free sector must be in range from 0 to 15, is %hhu.\n", track.sector[INFO_SEC].byte[OFFSET_FIRSTSECTOR]);
					fatal_error_found = 1;
				}
				else
				{
					if (verbose) printf ("First free sector isn't in range from 0 to 15, but ignoring it, because option --force was used.\n");
				}
			}
			
			/* header in Hobeta file have simple checksum, try verify it */
			if ((int16_t)hobetachecksum != (int16_t)(header.byte[HH_CRC] + 256*header.byte[HH_CRC+1]))
			{
				if (!force)
				{
					colored_warning ();
					printf ("Wrong checksum of header in the source Hobeta file.\n");
					fatal_error_found = 1;
				}
				else
				{
					if (verbose) printf ("Checksum of header doesn't match, but ignoring it, because option --force was used.\n");
				}
			}
			
			/* TRDOS image must have enough of free space */
			if (track.sector[INFO_SEC].byte[OFFSET_FILES] == 128)
			{
				colored_error ();
				printf ("Directory is full, no space to write any file.\n");
				fatal_error_found = 1;
			}
			if ((track.sector[INFO_SEC].byte[OFFSET_FREESECTORS] == 0) && (track.sector[INFO_SEC].byte[OFFSET_FREESECTORS+1] == 0))
			{
				colored_error ();
				printf ("Disk is full, no space to write any file.\n");
				fatal_error_found = 1;
			}
			if (header.byte[HH_SECTORS] > (track.sector[INFO_SEC].byte[OFFSET_FREESECTORS] + 256*track.sector[INFO_SEC].byte[OFFSET_FREESECTORS+1]))
			{
				colored_error ();
				printf ("Hobeta file is bigger (%hhu sectors) than remaining free space (%hhu sectors).\n", header.byte[HH_SECTORS], track.sector[INFO_SEC].byte[OFFSET_FREESECTORS] + 256*track.sector[INFO_SEC].byte[OFFSET_FREESECTORS+1]);
				fatal_error_found = 1;
			}
			
			/* check for duplicity, print warning if same (?) file already exists in disk image */
			dupfound = 0;
			for (trdosfile = 0; trdosfile < track.sector[INFO_SEC].byte[OFFSET_FILES]; trdosfile++)
			{
				/* check filename with extension, if match, compare length and start address */
				/* maybe option checksum can be used to compare content of files? */
				
				/* print filename with proper conversion of semigraphics and BASIC tokens */
				s = trdosfile / 16;		/* sector number in s - 16 bytes per item in directory */
				o = (trdosfile % 16)*16;	/* offset of filename in sector */
				
				/* print searched filenames when debugging */
				if (DEBUG) 
				{
					printf ("%hu,%hu\t[", s, o);
					for (i = 0; i < 8; i++) printzxchars(track.sector[s].byte[o+i]);
					printf (".");
					printzxchars(track.sector[s].byte[TR_EXTENSION]);
					printf ("]\n");
				}
				
				/* compare filenames */
				different = 0;
				for (i = 0; i < 9; i++)
				{
					if (header.byte[i] != track.sector[s].byte[o+i]) different = 1;
				}
				
				/* compare start and length */
				if (header.byte[HH_ADDRESS] != track.sector[s].byte[o+TR_ADDRESS]) different = 1;
				if (header.byte[HH_ADDRESS+1] != track.sector[s].byte[o+TR_ADDRESS+1]) different = 1;
				if (header.byte[HH_LENGTH] != track.sector[s].byte[o+TR_LENGTH]) different = 1;
				if (header.byte[HH_LENGTH+1] != track.sector[s].byte[o+TR_LENGTH+1]) different = 1;
				
				/* if after all checks no difference was found, then it is duplicity, source file is already in TRDOS image */
				if (different == 0)
				{
					/* first duplicity found is enough to break for cycle */
					dupfound = 1;
					break;
				}
			}
			
			/* warning if duplicate file found in directory. */
			if (dupfound == 1)
			{
				if (!force)
				{
					colored_warning ();
					printf ("File with same name \"");
					
					/* print which name on TRDOS image it is */
					for (i = 0; i < 8; i++) printzxchars(track.sector[s].byte[o+i]);
					printf (".");
					printzxchars(track.sector[s].byte[TR_EXTENSION]);
					printf ("\", and length already exists on TRDOS disk image.\n");
					
					if (verbose)
					{
						printf ("You can use option --force and copy it anyway or you can compare CRC32 first and check if files are really same. Utility lstrd can compute CRC32 for TRDOS image.\n");
					}
					fatal_error_found = 1;
				}
				else
				{
					if (verbose) printf ("File with same name is already on TRDOS image, but ignoring it, because option --force was used.\n");
				}
			}
			
			/* if any check failed, don't continue */
			if (fatal_error_found)
			{
				fclose (inputfile);
				fclose (outputfile);
				exit (EXIT_FAILURE);
			}
			
			/* in verbose mode print all details about source and target files */
			if (verbose)
			{
				/* source */
				printf ("Hobeta file: ");
				
				/* print filename with proper conversion of semigraphics and BASIC tokens */
				for (i = 0; i < 8; i++) printzxchars(header.byte[HH_FILENAME+i]);
				printf (".");
				printzxchars(header.byte[HH_EXTENSION]);
				
				/* numeric values */
				printf (" %hu,", (header.byte[HH_ADDRESS] + 256*header.byte[HH_ADDRESS+1]));
				printf (" %hu,", (header.byte[HH_LENGTH] + 256*header.byte[HH_LENGTH+1]));
				printf (" %hu sec.,", header.byte[HH_SECTORS]);
				printf (" checksum %hu\n", 0xFFFF & (int16_t)(header.byte[HH_CRC] + 256*header.byte[HH_CRC+1]));
				
				/* target */
				printf ("TRDOS image: %s\n", target_filename);
				printf ("Free %u sectors, Files %u, Deleted %u\n",track.sector[INFO_SEC].byte[OFFSET_FREESECTORS+1]*256+track.sector[INFO_SEC].byte[OFFSET_FREESECTORS], track.sector[INFO_SEC].byte[OFFSET_FILES], track.sector[INFO_SEC].byte[OFFSET_DELETEDFILES]);
			}
			
			/*
			   DATA SOUBORU
			   najít první volný sektor a stopu, zaseekovat v cílovém souboru a plnit jeden sektor za druhým dokud jsou 
			   data na vstupu
			   soubor hobeta je vždy dlouhé 256*n+17 bytů, kde n se mění podle počtu sektorů na TRDOSové disketě, které 
			   soubor obsazoval/obsadí => je možno načítat a zapisovat přímo po sektorech
			*/
			sector_number = track.sector[INFO_SEC].byte[OFFSET_FIRSTSECTOR];
			track_number = track.sector[INFO_SEC].byte[OFFSET_FIRSTTRACK];
			freesectors = track.sector[INFO_SEC].byte[OFFSET_FREESECTORS+1]*256+track.sector[INFO_SEC].byte[OFFSET_FREESECTORS];
			fseek (inputfile, 17, SEEK_SET);
			fseek (outputfile, sector_number*SEC_SIZE+track_number*TRK_SIZE*SEC_SIZE, SEEK_SET);
			while (fread (&sector, sizeof (trdos_sector), 1, inputfile) == 1)
			{
				if (verbose) printf ("[%u,%u] ", track_number, sector_number);
				sector_number++;
				if (sector_number>15)
				{
					sector_number = 0;
					track_number++;
				}
				freesectors--;
				if (freesectors < 0)
				{
					colored_error ();
					printf ("Out of disk space.\n");
					fclose (inputfile);
					fclose (outputfile);
					exit (EXIT_FAILURE);
				}
				
				if (!dryrun)
				{
					/* write sector */
					if (fwrite (&sector, sizeof (trdos_sector), 1, outputfile) != 1)
					{
						printf ("Writing of sector in %s failed: %s\n", target_filename, strerror(errno));
						fclose (inputfile);
						fclose (outputfile);
						exit (EXIT_FAILURE);
					}
				}
			}
			
			/* modifying TRDOS directory */
			fseek (outputfile, 0, SEEK_SET);
			dir_sector = track.sector[INFO_SEC].byte[OFFSET_FILES] / 16;	/* in which sector write filename of new file */
			directory_offset = (track.sector[INFO_SEC].byte[OFFSET_FILES] % 16)*16;	/* which item in directory will be modified */
			
			/* tvorba hlavičky souboru v adresáři (přenesu jméno+přípona+adresa+dĂŠlka - vše najednou, protože to jde) */
			for (i = 0; i < 13; i++) track.sector[dir_sector].byte[directory_offset+i] = header.byte[HH_FILENAME+i];
			track.sector[dir_sector].byte[directory_offset+13] = header.byte[HH_SECTORS];
			track.sector[dir_sector].byte[directory_offset+14] = track.sector[INFO_SEC].byte[OFFSET_FIRSTSECTOR];
			track.sector[dir_sector].byte[directory_offset+15] = track.sector[INFO_SEC].byte[OFFSET_FIRSTTRACK];
			
			/* provedení potřebných změn v systémovém sektoru TRDOSu (9. sektor číslo 8) */
			if (verbose) printf ("\nNew first track and sector %u, %u\n", track_number, sector_number); /* první volný sektor a track ... */
			track.sector[INFO_SEC].byte[OFFSET_FIRSTSECTOR] = sector_number;	/* zapiš čísla prvních volných sek. a tr. */
			track.sector[INFO_SEC].byte[OFFSET_FIRSTTRACK] = track_number;
			track.sector[INFO_SEC].byte[OFFSET_FREESECTORS] = freesectors % 256;	/* zapiš zbývající volné místo na disku */
			track.sector[INFO_SEC].byte[OFFSET_FREESECTORS+1] = freesectors / 256;
			track.sector[INFO_SEC].byte[OFFSET_FILES]++;				/* zvyš počet souborů o právě zapsaný */
			
			if (!dryrun)
			{
				/* write track 0 with modified directory & system sector in target TRDOS image */
				if (fwrite (&track, sizeof(trdos_track), 1, outputfile) != 1)
				{
					printf ("Writing of track 0 in %s failed.\n", target_filename);
					fclose (inputfile);
					fclose (outputfile);
					exit (EXIT_FAILURE);
				}
			}
			if (verbose) printf ("Free: %hu\nO.K.\n", freesectors);
		}
		else
		{
			colored_error ();
			printf ("Reading from file %s failed.\n", target_filename);
			fclose (inputfile);
			fclose (outputfile);
			exit (EXIT_FAILURE);
		}
	}
	else
	{
		colored_error ();
		printf ("Reading from file %s failed.\n", source_filename);
		fclose (inputfile);
		exit (EXIT_FAILURE);
	}
	/* data converted */
	return (EXIT_SUCCESS);
}


uint32_t compute_hobeta_crc32(char* source_filename, unsigned char mode)
{
	hobetahead header;
	unsigned char byte;
	unsigned int byte_counter;
	FILE *inputfile;
	uint32_t crc;
	uint16_t length;
	
	/* mode = 0 - will compute crc32 from all sectors, even if last one is not used whole */
	/* mode = 1 - will compute crc32 only from bytes of file */
	
	inputfile = fopen(source_filename, "rb");
	
	if (inputfile == NULL)
	{
		colored_error ();
		printf ("Can't open input file %s.\n", source_filename);
		exit (EXIT_FAILURE);
	}
	
	if (fread (&header, sizeof (hobetahead), 1, inputfile) == 1)
	{
		if (mode == 1)
		{
			if ((header.byte[HH_EXTENSION]) == 'b' || header.byte[HH_EXTENSION] == 'B')
			{
				/* length with variables for BASIC program */
				length = header.byte[HH_ADDRESS] + 256*header.byte[HH_ADDRESS+1];
			}
			else
			{
				/* binary files have only one length */
				length = header.byte[HH_LENGTH] + 256*header.byte[HH_LENGTH+1];
			}
		}
		else
		{
			length = 256*header.byte[HH_SECTORS];
		}
		
		if (DEBUG) printf ("CRC32 will be computed from %hu bytes.\n",length);
		
		/* initial CRC32 value */
		crc = 0xFFFFFFFF;
		
		/* CRC32 will be computed from data which are following after 17 bytes long header */
		fseek (inputfile, 17, SEEK_SET);
		
		byte_counter = 0;
		if (inputfile)
		{
			while (fread(&byte, 1, 1, inputfile) == 1)
			{
				if (mode == 1 && byte_counter == length)
				{
					if (DEBUG) printf ("\t\tCRC32 last byte was %i.\n", byte_counter);
					break;
				}

				crc = (crc >> 8) ^ crc32Lookup[(crc & 0xFF) ^ byte];
				byte_counter++;
			}
		}
		if (DEBUG && mode == 0) printf ("\t\tCRC32 last byte was %i.\n", (byte_counter - 1));
		
		crc = ~crc;	/* not minus! it is complement (bitwise NOT) */
	}
	else
	{
		colored_error ();
		printf ("Reading from file %s failed.\n", source_filename);
		fclose (inputfile);
		exit (EXIT_FAILURE);
	}

	fclose (inputfile);
	return (crc);
}


void version ()
{
	printf ("version %s \n", VERSION);
}


void help ()
{
	printf ("HOBETA2TRD is a utility for copying HOBETA files in to TRDOS disk image.\n");
	printf ("Usage: hobeta2trd [OPTIONS] [SOURCE] [TARGET]\n");
	printf ("       hobeta2trd  -i [OPTIONS] [SOURCE]\n");
	printf ("[SOURCE] must be any Hobeta file\n");
	printf ("[TARGET] must be TRDOS disk image\n");
	printf ("  -i, --info     will only display details about Hobeta file\n");
	printf ("  -V, --verbose  increase verbosity level\n");
	printf ("  -f, --force    forcefully copy Hobeta file into TRD image\n");
	printf ("  -C, --color    enable color output, it will highlight some values\n");
	printf ("  --dry-run      will simulate whole procedure except writing in TRD image\n");
	printf ("  -h, --help     this text\n");
	printf ("  -v, --version  version\n");
	printf ("\n");
	printf ("If you want to list content of TRD image, utility lstrd can help.\n");
	printf ("\n");
	printf ("My website https://cygnus.speccy.cz\n");
	printf ("Project website https://sourceforge.net/projects/zxspectrumutils\n");
}


int main(int argc, char *argv[])
{
	char *source_filename;
	char *target_filename;
	unsigned int counter;
	int switches;
	unsigned char correct;
	unsigned char hobetainfoonly;
	int returnvalue;
	uint32_t crc32;

	/* input file is parameter before last, output file is last parameter, other parameters must be correct */
	if (DEBUG) printf ("DEBUG: argc - %i\n", argc);
	
	hobetainfoonly = 0;
	verbose = 0;
	colors = 0;
	force = 0;

	switches = 1;
	for (counter = 1; counter < argc; counter++)
	{
		if (DEBUG) printf ("++ %i ++ %s \n",counter, argv[counter]);
		correct = 0;

		if (!strcmp(argv[counter], "-h") || (!strcmp(argv[counter], "--help")))
		{
			help ();
			switches++;
			correct = 1;
			return (EXIT_SUCCESS);
		}
		if (!strcmp(argv[counter], "-v") || (!strcmp(argv[counter], "--version")))
		{
			version ();
			switches++;
			correct = 1;
			return (EXIT_SUCCESS);
		}
		if (!strcmp(argv[counter], "-V") || (!strcmp(argv[counter], "--verbose")))
		{
			verbose = 1;
			switches++;
			correct = 1;
		}
		if (!strcmp(argv[counter], "-f") || (!strcmp(argv[counter], "--force")))
		{
			force = 1;
			switches++;
			correct = 1;
		}
		if (!strcmp(argv[counter], "--dry-run"))
		{
			/* simulate everything, but don't write in target file */
			dryrun = 1;
			switches++;
			correct = 1;
		}
		if (!strcmp(argv[counter], "-C") || (!strcmp(argv[counter], "--color")) || (!strcmp(argv[counter], "--colors")))
		{
			/* colors enabled */
			colors = 1;
			switches++;
			correct = 1;
		}
		if (!strcmp(argv[counter], "-i") || (!strcmp(argv[counter], "--info")))
		{
			hobetainfoonly = 1;
			switches++;
			correct = 1;
		}
		if ((correct == 0) && !((counter == (argc-1)) || (counter==(argc-2))))
		{
			/* parameter before last can be filename of source file */
			/* last parameter is filename of target TRD image */
			colored_error ();
			printf ("Unknown parameter - \"%s\"\n", argv[counter]);
			help ();
			exit (EXIT_FAILURE);
		}
	}

	if (hobetainfoonly)
	{
		/* Will display only details about Hobeta file and its content. */
		if (switches != argc-1)
		{
			colored_error ();
			printf ("Filename of source file is missing.\n");
			exit (EXIT_FAILURE);
		}
		else
		{
			source_filename = argv[argc-1];
			if (verbose) printf ("[SOURCE] %s \n", source_filename);
			
			returnvalue = hobetainfo (source_filename);
			if (verbose && returnvalue == 0)
			{
				printf ("crc32:    \t");
				crc32 = compute_hobeta_crc32 (source_filename,0);
				printf ("0x%08x\tcomputed from all data sectors\n", crc32);
				printf ("crc32:    \t");
				crc32 = compute_hobeta_crc32 (source_filename,1);
				printf ("0x%08x\tlimited to the length of file\n", crc32);
			}
			else
				return (returnvalue);
		}
	}
	else
	{
		if (switches != argc-2)
		{
			colored_error ();
			printf ("At least one of filenames is missing.\n");
			exit (EXIT_FAILURE);
		}
		else
		{
			source_filename = argv[argc-2];
			target_filename = argv[argc-1];
			if (verbose) printf ("[SOURCE] %s -> [TARGET] %s \n", source_filename, target_filename);
			
			return (convert (source_filename, target_filename));
		}
	}
}
